



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
nosetests --where=tests --ignore-files=citeproc-test.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
python tests/citeproc-test.py --no-update
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
